@@ -14,144 +14,194 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 # ==============================================================================
-from __future__ import absolute_import
-from __future__ import division
-from __future__ import print_function
-
 import argparse
 import logging
 import os
 import sys
+from typing import List
 
 import numpy as np
 
-from ludwig.constants import TEST, TRAINING, VALIDATION, FULL
+from ludwig.api import LudwigModel
+from ludwig.constants import FULL, TEST, TRAINING, VALIDATION
 from ludwig.contrib import contrib_command
-from ludwig.data.preprocessing import preprocess_for_prediction
 from ludwig.globals import LUDWIG_VERSION
-from ludwig.globals import TRAIN_SET_METADATA_FILE_NAME
-from ludwig.models.model import load_model_and_definition
-from ludwig.utils.misc import find_non_existing_dir_by_adding_suffix
-from ludwig.utils.print_utils import logging_level_registry
-from ludwig.utils.print_utils import print_boxed
-from ludwig.utils.print_utils import print_ludwig
+from ludwig.utils.print_utils import (logging_level_registry, print_boxed,
+                                      print_ludwig)
 from ludwig.utils.strings_utils import make_safe_filename
 
 logger = logging.getLogger(__name__)
 
 
 def collect_activations(
-        model_path,
-        tensors,
-        data_csv=None,
-        data_hdf5=None,
-        split=TEST,
-        batch_size=128,
-        output_directory='results',
-        gpus=None,
-        gpu_fraction=1.0,
-        debug=False,
+        model_path: str,
+        layers: List[str],
+        dataset: str,
+        data_format: str = None,
+        split: str = FULL,
+        batch_size: int = 128,
+        output_directory: str = 'results',
+        gpus: List[str] = None,
+        gpu_memory_limit: int =None,
+        allow_parallel_threads: bool = True,
+        use_horovod: bool = None,
+        debug: bool = False,
         **kwargs
-):
-    """Uses the pretrained model to collect the tensors corresponding to a
+) -> List[str]:
+    """
+    Uses the pretrained model to collect the tensors corresponding to a
     datapoint in the dataset. Saves the tensors to the experiment directory
 
-    :param model_path: Is the model from which the tensors will be collected
-    :param tensors: List contaning the names of the tensors to collect
-    :param data_csv: The CSV filepath which contains the datapoints from which
-           the tensors are collected
-    :param data_hdf5: The HDF5 file path if the CSV file path does not exist,
-           an alternative source of providing the data to the model
-    :param split: Split type
-    :param batch_size: Batch size
-    :param output_directory: Output directory
-    :param gpus: The total number of GPUs that the model intends to use
-    :param gpu_fraction: The fraction of each GPU that the model intends on
-           using
-    :param debug: To step through the stack traces and find possible errors
-    :returns: None
+    # Inputs
 
+    :param model_path: (str) filepath to pre-trained model.
+    :param layers: (List[str]) list of strings for layer names in the model
+        to collect activations.
+    :param dataset: (str) source
+        containing the data to make predictions.
+    :param data_format: (str, default: `None`) format to interpret data
+        sources. Will be inferred automatically if not specified.  Valid
+        formats are `'auto'`, `'csv'`, `'excel'`, `'feather'`,
+        `'fwf'`, `'hdf5'` (cache file produced during previous training),
+        `'html'` (file containing a single HTML `<table>`), `'json'`, `'jsonl'`,
+        `'parquet'`, `'pickle'` (pickled Pandas DataFrame), `'sas'`, `'spss'`,
+        `'stata'`, `'tsv'`.
+    :param split: (str, default: `full`) split on which
+        to perform predictions. Valid values are `'training'`, `'validation'`,
+        `'test'` and `'full'`.
+    :param batch_size: (int, default `128`) size of batches for processing.
+    :param output_directory: (str, default: `'results'`) the directory that
+        will contain the training statistics, TensorBoard logs, the saved
+        model and the training progress files.
+    :param gpus: (list, default: `None`) list of GPUs that are available
+        for training.
+    :param gpu_memory_limit: (int, default: `None`) maximum memory in MB to
+        allocate per GPU device.
+    :param allow_parallel_threads: (bool, default: `True`) allow TensorFlow
+        to use multithreading parallelism to improve performance at
+        the cost of determinism.
+    :param debug: (bool, default: `False) if `True` turns on `tfdbg` with
+        `inf_or_nan` checks.
+
+    # Return
+
+    :return: (List[str]) list of filepath to `*.npy` files containing
+        the activations.
     """
-    # setup directories and file names
-    experiment_dir_name = find_non_existing_dir_by_adding_suffix(output_directory)
-
-    logger.info('Dataset path: {}'.format(
-        data_csv if data_csv is not None else data_hdf5)
-    )
+    logger.info('Dataset path: {}'.format(dataset)
+                )
     logger.info('Model path: {}'.format(model_path))
-    logger.info('Output path: {}'.format(experiment_dir_name))
+    logger.info('Output path: {}'.format(output_directory))
     logger.info('\n')
 
-    train_set_metadata_fp = os.path.join(
+    model = LudwigModel.load(
         model_path,
-        TRAIN_SET_METADATA_FILE_NAME
+        gpus=gpus,
+        gpu_memory_limit=gpu_memory_limit,
+        allow_parallel_threads=allow_parallel_threads,
+        use_horovod=use_horovod
     )
 
-    # preprocessing
-    dataset, train_set_metadata = preprocess_for_prediction(
-        model_path,
-        split,
-        data_csv,
-        data_hdf5,
-        train_set_metadata_fp
-    )
-
-    model, model_definition = load_model_and_definition(model_path)
-
     # collect activations
     print_boxed('COLLECT ACTIVATIONS')
     collected_tensors = model.collect_activations(
+        layers,
         dataset,
-        tensors,
-        batch_size,
-        gpus=gpus,
-        gpu_fraction=gpu_fraction
+        data_format=data_format,
+        split=split,
+        batch_size=batch_size,
+        debug=debug
     )
 
-    model.close_session()
-
     # saving
-    os.makedirs(experiment_dir_name)
-    save_tensors(collected_tensors, experiment_dir_name)
+    os.makedirs(output_directory, exist_ok=True)
+    saved_filenames = save_tensors(collected_tensors, output_directory)
 
-    logger.info('Saved to: {0}'.format(experiment_dir_name))
+    logger.info('Saved to: {0}'.format(output_directory))
+    return saved_filenames
 
 
 def collect_weights(
-        model_path,
-        tensors,
-        output_directory='results',
-        debug=False,
+        model_path: str,
+        tensors: List[str],
+        output_directory: str = 'results',
+        debug: bool = False,
         **kwargs
-):
-    # setup directories and file names
-    experiment_dir_name = find_non_existing_dir_by_adding_suffix(output_directory)
+) -> List[str]:
+    """
+    Loads a pretrained model and collects weights.
 
+    # Inputs
+    :param model_path: (str) filepath to pre-trained model.
+    :param tensors: (list, default: `None`) List of tensor names to collect
+        weights
+    :param output_directory: (str, default: `'results'`) the directory where
+        collected weights will be stored.
+    :param debug: (bool, default: `False) if `True` turns on `tfdbg` with
+        `inf_or_nan` checks.
+
+    # Return
+
+    :return: (List[str]) list of filepath to `*.npy` files containing
+        the weights.
+    """
     logger.info('Model path: {}'.format(model_path))
-    logger.info('Output path: {}'.format(experiment_dir_name))
+    logger.info('Output path: {}'.format(output_directory))
     logger.info('\n')
 
-    model, model_definition = load_model_and_definition(model_path)
+    model = LudwigModel.load(model_path)
 
     # collect weights
     print_boxed('COLLECT WEIGHTS')
     collected_tensors = model.collect_weights(tensors)
-    model.close_session()
 
     # saving
-    os.makedirs(experiment_dir_name)
-    save_tensors(collected_tensors, experiment_dir_name)
+    os.makedirs(output_directory, exist_ok=True)
+    saved_filenames = save_tensors(collected_tensors, output_directory)
 
-    logger.info('Saved to: {0}'.format(experiment_dir_name))
+    logger.info('Saved to: {0}'.format(output_directory))
+    return saved_filenames
 
 
-def save_tensors(collected_tensors, experiment_dir_name):
-    for tensor_name, tensor_values in collected_tensors.items():
+def save_tensors(collected_tensors, output_directory):
+    filenames = []
+    for tensor_name, tensor_value in collected_tensors:
         np_filename = os.path.join(
-            experiment_dir_name,
+            output_directory,
             make_safe_filename(tensor_name) + '.npy'
         )
-        np.save(np_filename, tensor_values)
+        np.save(np_filename, tensor_value.numpy())
+        filenames.append(np_filename)
+    return filenames
+
+
+def print_model_summary(
+        model_path: str,
+        **kwargs
+) -> None:
+    """
+    Loads a pretrained model and prints names of weights and layers activations.
+
+    # Inputs
+    :param model_path: (str) filepath to pre-trained model.
+
+    # Return
+    :return: (`None`)
+    """
+    model = LudwigModel.load(model_path)
+    collected_tensors = model.collect_weights()
+    names = [name for name, w in collected_tensors]
+
+    keras_model = model.model.get_connected_model(training=False)
+    keras_model.summary()
+
+    print('\nLayers:\n')
+    for layer in keras_model.layers:
+        print(layer.name)
+
+    print('\nWeights:\n')
+    for name in names:
+        print(name)
 
 
 def cli_collect_activations(sys_argv):
@@ -184,16 +234,25 @@ def cli_collect_activations(sys_argv):
     # ---------------
     # Data parameters
     # ---------------
-    group = parser.add_mutually_exclusive_group(required=True)
-    group.add_argument('--data_csv', help='input data CSV file')
-    group.add_argument('--data_hdf5', help='input data HDF5 file')
-
+    parser.add_argument(
+        '--dataset',
+        help='input data file path',
+        required=True
+    )
+    parser.add_argument(
+        '--data_format',
+        help='format of the input data',
+        default='auto',
+        choices=['auto', 'csv', 'excel', 'feather', 'fwf', 'hdf5',
+                 'html' 'tables', 'json', 'jsonl', 'parquet', 'pickle', 'sas',
+                 'spss', 'stata', 'tsv']
+    )
     parser.add_argument(
         '-s',
         '--split',
-        default=TEST,
+        default=FULL,
         choices=[TRAINING, VALIDATION, TEST, FULL],
-        help='the split to test the model on'
+        help='the split to obtain the model activations from'
     )
 
     # ----------------
@@ -206,13 +265,15 @@ def cli_collect_activations(sys_argv):
         required=True
     )
     parser.add_argument(
-        '-t',
-        '--tensors',
+        '-lyr',
+        '--layers',
         help='tensors to collect',
         nargs='+',
         required=True
     )
 
+
+
     # -------------------------
     # Output results parameters
     # -------------------------
@@ -246,11 +307,25 @@ def cli_collect_activations(sys_argv):
         help='list of gpu to use'
     )
     parser.add_argument(
-        '-gf',
-        '--gpu_fraction',
-        type=float,
-        default=1.0,
-        help='fraction of gpu memory to initialize the process with'
+        '-gml',
+        '--gpu_memory_limit',
+        type=int,
+        default=None,
+        help='maximum memory in MB to allocate per GPU device'
+    )
+    parser.add_argument(
+        '-dpt',
+        '--disable_parallel_threads',
+        action='store_false',
+        dest='allow_parallel_threads',
+        help='disable TensorFlow from using multithreading for reproducibility'
+    )
+    parser.add_argument(
+        '-uh',
+        '--use_horovod',
+        action='store_true',
+        default=None,
+        help='uses horovod for distributed training'
     )
     parser.add_argument(
         '-dbg',
@@ -269,8 +344,9 @@ def cli_collect_activations(sys_argv):
 
     args = parser.parse_args(sys_argv)
 
+    args.logging_level = logging_level_registry[args.logging_level]
     logging.getLogger('ludwig').setLevel(
-        logging_level_registry[args.logging_level]
+        args.logging_level
     )
     global logger
     logger = logging.getLogger('ludwig.collect')
@@ -344,8 +420,9 @@ def cli_collect_weights(sys_argv):
 
     args = parser.parse_args(sys_argv)
 
+    args.logging_level = logging_level_registry[args.logging_level]
     logging.getLogger('ludwig').setLevel(
-        logging_level_registry[args.logging_level]
+        args.logging_level
     )
     global logger
     logger = logging.getLogger('ludwig.collect')
@@ -355,6 +432,55 @@ def cli_collect_weights(sys_argv):
     collect_weights(**vars(args))
 
 
+def cli_collect_summary(sys_argv):
+    """Command Line Interface to collecting a summary of the model layers and weights.
+    --m: Input model that is necessary to collect to the tensors, this is a
+         required *option*
+    --v: Verbose: Defines the logging level that the user will be exposed to
+    """
+    parser = argparse.ArgumentParser(
+        description='This script loads a pretrained model '
+                    'and prints names of weights and layers activations '
+                    'to use with other collect commands',
+        prog='ludwig collect_summary',
+        usage='%(prog)s [options]'
+    )
+
+    # ----------------
+    # Model parameters
+    # ----------------
+    parser.add_argument(
+        '-m',
+        '--model_path',
+        help='model to load',
+        required=True
+    )
+
+    # ------------------
+    # Runtime parameters
+    # ------------------
+    parser.add_argument(
+        '-l',
+        '--logging_level',
+        default='info',
+        help='the level of logging to use',
+        choices=['critical', 'error', 'warning', 'info', 'debug', 'notset']
+    )
+
+    args = parser.parse_args(sys_argv)
+
+    args.logging_level = logging_level_registry[args.logging_level]
+    logging.getLogger('ludwig').setLevel(
+        args.logging_level
+    )
+    global logger
+    logger = logging.getLogger('ludwig.collect')
+
+    print_ludwig('Collect Summary', LUDWIG_VERSION)
+
+    print_model_summary(**vars(args))
+
+
 if __name__ == '__main__':
     if len(sys.argv) > 1:
         if sys.argv[1] == 'activations':
@@ -363,6 +489,9 @@ if __name__ == '__main__':
         elif sys.argv[1] == 'weights':
             contrib_command("collect_weights", *sys.argv)
             cli_collect_weights(sys.argv[2:])
+        elif sys.argv[1] == 'names':
+            contrib_command("collect_summary", *sys.argv)
+            cli_collect_summary(sys.argv[2:])
         else:
             print('Unrecognized command')
     else:
