@@ -21,11 +21,12 @@ import sys
 import numpy as np
 import tensorflow as tf
 
-from ludwig.constants import AUDIO, BACKFILL
-from ludwig.features.base_feature import BaseFeature
+from ludwig.constants import AUDIO, BACKFILL, TIED, TYPE, COLUMN, PROC_COLUMN, \
+    PREPROCESSING, NAME
+from ludwig.encoders.sequence_encoders import StackedCNN, ParallelCNN, \
+    StackedParallelCNN, StackedRNN, SequencePassthroughEncoder, StackedCNNRNN
 from ludwig.features.sequence_feature import SequenceInputFeature
-from ludwig.utils.audio_utils import calculate_incr_mean
-from ludwig.utils.audio_utils import calculate_incr_var
+from ludwig.utils.audio_utils import calculate_mean, calculate_var
 from ludwig.utils.audio_utils import get_fbank
 from ludwig.utils.audio_utils import get_group_delay
 from ludwig.utils.audio_utils import get_length_in_samp
@@ -34,16 +35,14 @@ from ludwig.utils.audio_utils import get_non_symmetric_length
 from ludwig.utils.audio_utils import get_phase_stft_magnitude
 from ludwig.utils.audio_utils import get_stft_magnitude
 from ludwig.utils.data_utils import get_abs_path
-from ludwig.utils.misc import set_default_value
-from ludwig.utils.misc import set_default_values
+from ludwig.utils.misc_utils import set_default_value
+from ludwig.utils.misc_utils import set_default_values
 
 logger = logging.getLogger(__name__)
 
 
-class AudioBaseFeature(BaseFeature):
-    def __init__(self, feature):
-        super().__init__(feature)
-        self.type = AUDIO
+class AudioFeatureMixin(object):
+    type = AUDIO
 
     preprocessing_defaults = {
         'audio_file_length_limit_in_s': 7.5,
@@ -52,12 +51,16 @@ class AudioBaseFeature(BaseFeature):
         'padding_value': 0,
         'norm': None,
         'audio_feature': {
-            'type': 'raw',
+            TYPE: 'raw',
         }
     }
 
     @staticmethod
-    def get_feature_meta(column, preprocessing_parameters):
+    def cast_column(feature, dataset_df, backend):
+        return dataset_df
+
+    @staticmethod
+    def get_feature_meta(column, preprocessing_parameters, backend):
         try:
             import soundfile
         except ImportError:
@@ -69,14 +72,14 @@ class AudioBaseFeature(BaseFeature):
             sys.exit(-1)
 
         audio_feature_dict = preprocessing_parameters['audio_feature']
-        first_audio_file_path = column[0]
+        first_audio_file_path = column.head(1)[0]
         _, sampling_rate_in_hz = soundfile.read(first_audio_file_path)
 
-        feature_dim = AudioBaseFeature._get_feature_dim(audio_feature_dict,
-                                                        sampling_rate_in_hz)
+        feature_dim = AudioFeatureMixin._get_feature_dim(audio_feature_dict,
+                                                         sampling_rate_in_hz)
         audio_file_length_limit_in_s = preprocessing_parameters[
             'audio_file_length_limit_in_s']
-        max_length = AudioBaseFeature._get_max_length_feature(
+        max_length = AudioFeatureMixin._get_max_length_feature(
             audio_feature_dict, sampling_rate_in_hz,
             audio_file_length_limit_in_s)
         return {
@@ -87,7 +90,7 @@ class AudioBaseFeature(BaseFeature):
 
     @staticmethod
     def _get_feature_dim(audio_feature_dict, sampling_rate_in_hz):
-        feature_type = audio_feature_dict['type']
+        feature_type = audio_feature_dict[TYPE]
 
         if feature_type == 'raw':
             feature_dim = 1
@@ -107,16 +110,17 @@ class AudioBaseFeature(BaseFeature):
         return feature_dim
 
     @staticmethod
-    def _read_audio_and_transform_to_feature(filepath, audio_feature_dict,
-                                             feature_dim, max_length,
-                                             padding_value, normalization_type,
-                                             audio_stats):
-        """
-        :param filepath: path to the audio
-        :param audio_feature_dict: dictionary describing audio feature see default
-        :param feature_dim: dimension of each feature frame
-        :param max_length: max audio length defined by user in samples
-        """
+    def _process_in_memory(
+            column,
+            src_path,
+            audio_feature_dict,
+            feature_dim,
+            max_length,
+            padding_value,
+            normalization_type,
+            audio_file_length_limit_in_s,
+            backend
+    ):
         try:
             import soundfile
         except ImportError:
@@ -127,17 +131,66 @@ class AudioBaseFeature(BaseFeature):
             )
             sys.exit(-1)
 
-        feature_type = audio_feature_dict['type']
-        audio, sampling_rate_in_hz = soundfile.read(filepath)
-        AudioBaseFeature._update(audio_stats, audio, sampling_rate_in_hz)
+        def read_audio(path):
+            filepath = get_abs_path(src_path, path)
+            return soundfile.read(filepath)
 
+        df_engine = backend.df_engine
+        raw_audio = df_engine.map_objects(column, read_audio)
+        processed_audio = df_engine.map_objects(
+            raw_audio,
+            lambda row: AudioFeatureMixin._transform_to_feature(
+                audio=row[0],
+                sampling_rate_in_hz=row[1],
+                audio_feature_dict=audio_feature_dict,
+                feature_dim=feature_dim,
+                max_length=max_length,
+                padding_value=padding_value,
+                normalization_type=normalization_type
+            )
+        )
+
+        audio_stats = df_engine.map_objects(
+            raw_audio,
+            lambda row: AudioFeatureMixin._get_stats(
+                audio=row[0],
+                sampling_rate_in_hz=row[1],
+                max_length_in_s=audio_file_length_limit_in_s,
+            )
+        )
+
+        def reduce(series):
+            merged_stats = None
+            for audio_stats in series:
+                if merged_stats is None:
+                    merged_stats = audio_stats.copy()
+                else:
+                    AudioFeatureMixin._merge_stats(merged_stats, audio_stats)
+            return merged_stats
+
+        merged_stats = df_engine.reduce_objects(audio_stats, reduce)
+        merged_stats['mean'] = calculate_mean(merged_stats['sum'], merged_stats['count'])
+        merged_stats['var'] = calculate_var(merged_stats['sum'], merged_stats['sum2'], merged_stats['count'])
+        return processed_audio, merged_stats
+
+    @staticmethod
+    def _transform_to_feature(
+            audio,
+            sampling_rate_in_hz,
+            audio_feature_dict,
+            feature_dim,
+            max_length,
+            padding_value,
+            normalization_type
+    ):
+        feature_type = audio_feature_dict[TYPE]
         if feature_type == 'raw':
             audio_feature = np.expand_dims(audio, axis=-1)
         elif feature_type in ['stft', 'stft_phase', 'group_delay', 'fbank']:
             audio_feature = np.transpose(
-                AudioBaseFeature._get_2D_feature(audio, feature_type,
-                                                 audio_feature_dict,
-                                                 sampling_rate_in_hz))
+                AudioFeatureMixin._get_2D_feature(audio, feature_type,
+                                                  audio_feature_dict,
+                                                  sampling_rate_in_hz))
         else:
             raise ValueError('{} is not recognized.'.format(feature_type))
 
@@ -151,7 +204,8 @@ class AudioBaseFeature(BaseFeature):
 
         feature_length = audio_feature.shape[0]
         broadcast_feature_length = min(feature_length, max_length)
-        audio_feature_padded = np.full((max_length, feature_dim), padding_value,
+        audio_feature_padded = np.full((max_length, feature_dim),
+                                       padding_value,
                                        dtype=np.float32)
         audio_feature_padded[:broadcast_feature_length, :] = audio_feature[
                                                              :max_length, :]
@@ -159,21 +213,25 @@ class AudioBaseFeature(BaseFeature):
         return audio_feature_padded
 
     @staticmethod
-    def _update(audio_stats, audio, sampling_rate_in_hz):
+    def _get_stats(audio, sampling_rate_in_hz, max_length_in_s):
         audio_length_in_s = audio.shape[-1] / float(sampling_rate_in_hz)
-        audio_stats['count'] += 1
-        mean = ((audio_stats['count'] - 1) * audio_stats[
-            'mean'] + audio_length_in_s) / float(audio_stats['count'])
-        mean = calculate_incr_mean(audio_stats['count'], audio_stats['mean'],
-                                   audio_length_in_s)
-        var = calculate_incr_var(audio_stats['var'], audio_stats['mean'], mean,
-                                 audio_length_in_s)
-        audio_stats['mean'] = mean
-        audio_stats['var'] = var
-        audio_stats['max'] = max(audio_stats['max'], audio_length_in_s)
-        audio_stats['min'] = min(audio_stats['min'], audio_length_in_s)
-        if audio_length_in_s > audio_stats['max_length_in_s']:
-            audio_stats['cropped'] += 1
+        return {
+            'count': 1,
+            'sum': audio_length_in_s,
+            'sum2': audio_length_in_s * audio_length_in_s,
+            'min': audio_length_in_s,
+            'max': audio_length_in_s,
+            'cropped': 1 if audio_length_in_s > max_length_in_s else 0
+        }
+
+    @staticmethod
+    def _merge_stats(merged_stats, audio_stats):
+        merged_stats['count'] += audio_stats['count']
+        merged_stats['sum'] += audio_stats['sum']
+        merged_stats['sum2'] += audio_stats['sum2']
+        merged_stats['min'] = min(merged_stats['min'], audio_stats['min'])
+        merged_stats['max'] = max(merged_stats['max'], audio_stats['max'])
+        merged_stats['cropped'] += audio_stats['cropped']
 
     @staticmethod
     def _get_2D_feature(audio, feature_type, audio_feature_dict,
@@ -221,10 +279,11 @@ class AudioBaseFeature(BaseFeature):
     @staticmethod
     def add_feature_data(
             feature,
-            dataset_df,
-            data,
+            input_df,
+            proc_df,
             metadata,
-            preprocessing_parameters
+            preprocessing_parameters,
+            backend
     ):
         set_default_value(
             feature['preprocessing'],
@@ -232,31 +291,35 @@ class AudioBaseFeature(BaseFeature):
             preprocessing_parameters['in_memory']
         )
 
-        if not 'audio_feature' in preprocessing_parameters:
+        if 'audio_feature' not in preprocessing_parameters:
             raise ValueError(
                 'audio_feature dictionary has to be present in preprocessing '
                 'for audio.')
-        if not 'type' in preprocessing_parameters['audio_feature']:
+        if TYPE not in preprocessing_parameters['audio_feature']:
             raise ValueError(
                 'type has to be present in audio_feature dictionary '
                 'for audio.')
 
-        csv_path = None
-        if hasattr(dataset_df, 'csv'):
-            csv_path = os.path.dirname(os.path.abspath(dataset_df.csv))
-        if (csv_path is None and
-                not os.path.isabs(dataset_df[feature['name']][0])):
-            raise ValueError(
-                'Audio file paths must be absolute'
-            )
+        name = feature[NAME]
+        column = feature[COLUMN]
+        proc_column = feature[PROC_COLUMN]
 
-        num_audio_utterances = len(dataset_df)
+        src_path = None
+        # this is not super nice, but works both and DFs and lists
+        first_path = '.'
+        for first_path in input_df[column]:
+            break
+        if hasattr(input_df, 'src'):
+            src_path = os.path.dirname(os.path.abspath(input_df.src))
+        if src_path is None and not os.path.isabs(first_path):
+            raise ValueError('Audio file paths must be absolute')
+
+        num_audio_utterances = len(input_df)
         padding_value = preprocessing_parameters['padding_value']
         normalization_type = preprocessing_parameters['norm']
-        feature_name = feature['name']
 
-        feature_dim = metadata[feature_name]['feature_dim']
-        max_length = metadata[feature_name]['max_length']
+        feature_dim = metadata[name]['feature_dim']
+        max_length = metadata[name]['max_length']
         audio_feature_dict = preprocessing_parameters['audio_feature']
         audio_file_length_limit_in_s = preprocessing_parameters[
             'audio_file_length_limit_in_s']
@@ -264,50 +327,42 @@ class AudioBaseFeature(BaseFeature):
         if num_audio_utterances == 0:
             raise ValueError(
                 'There are no audio files in the dataset provided.')
-        audio_stats = {
-            'count': 0,
-            'mean': 0,
-            'var': 0,
-            'std': 0,
-            'max': 0,
-            'min': float('inf'),
-            'cropped': 0,
-            'max_length_in_s': audio_file_length_limit_in_s
-        }
 
-        if feature['preprocessing']['in_memory']:
-            data[feature['name']] = np.empty(
-                (num_audio_utterances, max_length, feature_dim),
-                dtype=np.float32
+        if feature[PREPROCESSING]['in_memory']:
+            audio_features, audio_stats = AudioFeatureMixin._process_in_memory(
+                input_df[feature[NAME]],
+                src_path,
+                audio_feature_dict,
+                feature_dim,
+                max_length,
+                padding_value,
+                normalization_type,
+                audio_file_length_limit_in_s,
+                backend
             )
-            for i in range(len(dataset_df)):
-                filepath = get_abs_path(
-                    csv_path,
-                    dataset_df[feature['name']][i]
-                )
-                audio_feature = AudioBaseFeature._read_audio_and_transform_to_feature(
-                    filepath, audio_feature_dict, feature_dim, max_length,
-                    padding_value, normalization_type, audio_stats
-                )
-
-                data[feature['name']][i, :, :] = audio_feature
+            proc_df[proc_column] = audio_features
 
             audio_stats['std'] = np.sqrt(
                 audio_stats['var'] / float(audio_stats['count']))
-            print_statistics = """
-            {} audio files loaded.
-            Statistics of audio file lengths:
-            - mean: {:.4f}
-            - std: {:.4f}
-            - max: {:.4f}
-            - min: {:.4f}
-            - cropped audio_files: {}
-            Max length was given as {}.
-            """.format(audio_stats['count'], audio_stats['mean'],
-                       audio_stats['std'], audio_stats['max'],
-                       audio_stats['min'], audio_stats['cropped'],
-                       audio_stats['max_length_in_s'])
-            print(print_statistics)
+            print_statistics = (
+                "{} audio files loaded.\n"
+                "Statistics of audio file lengths:\n"
+                "- mean: {:.4f}\n"
+                "- std: {:.4f}\n"
+                "- max: {:.4f}\n"
+                "- min: {:.4f}\n"
+                "- cropped audio_files: {}\n"
+                "Max length was given as {}s"
+            ).format(
+                audio_stats['count'], audio_stats['mean'],
+                audio_stats['std'], audio_stats['max'],
+                audio_stats['min'], audio_stats['cropped'],
+                audio_file_length_limit_in_s)
+            logger.debug(print_statistics)
+        else:
+            backend.check_lazy_load_supported(feature)
+
+        return proc_df
 
     @staticmethod
     def _get_max_length_feature(
@@ -315,7 +370,7 @@ class AudioBaseFeature(BaseFeature):
             sampling_rate_in_hz,
             audio_length_limit_in_s
     ):
-        feature_type = audio_feature_dict['type']
+        feature_type = audio_feature_dict[TYPE]
         audio_length_limit_in_samp = (
                 audio_length_limit_in_s * sampling_rate_in_hz
         )
@@ -341,54 +396,48 @@ class AudioBaseFeature(BaseFeature):
             raise ValueError('{} is not recognized.'.format(feature_type))
 
 
-class AudioInputFeature(AudioBaseFeature, SequenceInputFeature):
-    def __init__(self, feature):
-        super().__init__(feature)
-        self.type = AUDIO
-        self.length = None
-        self.embedding_size = None
-        self.overwrite_defaults(feature)
+class AudioInputFeature(AudioFeatureMixin, SequenceInputFeature):
+    encoder = 'parallel_cnn'
+    max_sequence_length = None
+    embedding_size = None
+
+    def __init__(self, feature, encoder_obj=None):
+        super().__init__(feature, encoder_obj=encoder_obj)
         if not self.embedding_size:
             raise ValueError(
                 'embedding_size has to be defined - '
-                'check "update_model_definition_with_metadata()"')
-        if not self.length:
+                'check "update_config_with_metadata()"')
+        if not self.max_sequence_length:
             raise ValueError(
-                'length has to be defined - '
-                'check "update_model_definition_with_metadata()"')
+                'max_sequence_length has to be defined - '
+                'check "update_config_with_metadata()"')
 
-    def _get_input_placeholder(self):
-        return tf.compat.v1.placeholder(
-            tf.float32, shape=[None, self.length, self.embedding_size],
-            name='{}_placeholder'.format(self.name)
+    def call(self, inputs, training=None, mask=None):
+        assert isinstance(inputs, tf.Tensor)
+        assert inputs.dtype == tf.float32
+        assert len(inputs.shape) == 3
+
+        encoder_output = self.encoder_obj(
+            inputs, training=training, mask=mask
         )
 
-    def build_input(
-            self,
-            regularizer,
-            dropout_rate,
-            is_training=False,
-            **kwargs
-    ):
-        placeholder = self._get_input_placeholder()
-        logger.debug('  placeholder: {0}'.format(placeholder))
+        return encoder_output
 
-        return self.build_sequence_input(
-            placeholder,
-            self.encoder_obj,
-            regularizer,
-            dropout_rate,
-            is_training
-        )
+    @classmethod
+    def get_input_dtype(cls):
+        return tf.float32
+
+    def get_input_shape(self):
+        return self.max_sequence_length, self.embedding_size
 
     @staticmethod
-    def update_model_definition_with_metadata(
+    def update_config_with_metadata(
             input_feature,
             feature_metadata,
             *args,
             **kwargs
     ):
-        input_feature['length'] = feature_metadata['max_length']
+        input_feature['max_sequence_length'] = feature_metadata['max_length']
         input_feature['embedding_size'] = feature_metadata['feature_dim']
         input_feature['should_embed'] = False
 
@@ -397,7 +446,20 @@ class AudioInputFeature(AudioBaseFeature, SequenceInputFeature):
         set_default_values(
             input_feature,
             {
-                'tied_weights': None,
+                TIED: None,
                 'preprocessing': {}
             }
         )
+
+    encoder_registry = {
+        'stacked_cnn': StackedCNN,
+        'parallel_cnn': ParallelCNN,
+        'stacked_parallel_cnn': StackedParallelCNN,
+        'rnn': StackedRNN,
+        'cnnrnn': StackedCNNRNN,
+        'passthrough': SequencePassthroughEncoder,
+        'null': SequencePassthroughEncoder,
+        'none': SequencePassthroughEncoder,
+        'None': SequencePassthroughEncoder,
+        None: SequencePassthroughEncoder
+    }
