@@ -453,14 +453,15 @@ class ParallelExecutor(HyperoptExecutor):
                             data_format=data_format,
                             experiment_name=f'{experiment_name}_{trial_id}',
                             model_name=model_name,
-                            # model_load_pat=model_load_path,
+                            # model_load_path=model_load_path,
                             # model_resume_path=model_resume_path,
                             skip_save_training_description=skip_save_training_description,
                             skip_save_training_statistics=skip_save_training_statistics,
                             skip_save_model=skip_save_model,
                             skip_save_progress=skip_save_progress,
                             skip_save_log=skip_save_log,
-                            skip_save_processed_input=skip_save_processed_input,
+                            # needed because of concurrent HDF5 writes
+                            skip_save_processed_input=True,
                             skip_save_unprocessed_output=skip_save_unprocessed_output,
                             skip_save_predictions=skip_save_predictions,
                             skip_save_eval_stats=skip_save_eval_stats,
