@@ -16,6 +16,7 @@
 # ==============================================================================
 import collections
 import csv
+import functools
 import json
 import logging
 import os.path
@@ -29,10 +30,47 @@ import pandas as pd
 from pandas.errors import ParserError
 from sklearn.model_selection import KFold
 
-from ludwig.constants import SPLIT
+from ludwig.constants import PREPROCESSING, SPLIT, PROC_COLUMN
+from ludwig.globals import (MODEL_HYPERPARAMETERS_FILE_NAME,
+                            MODEL_WEIGHTS_FILE_NAME,
+                            TRAIN_SET_METADATA_FILE_NAME)
 
 logger = logging.getLogger(__name__)
 
+DATASET_SPLIT_URL = 'dataset_{}_fp'
+DATA_PROCESSED_CACHE_DIR = 'data_processed_cache_dir'
+DATA_TRAIN_HDF5_FP = 'data_train_hdf5_fp'
+HDF5_COLUMNS_KEY = 'columns'
+DICT_FORMATS = {'dict', 'dictionary', dict}
+DATAFRAME_FORMATS = {'dataframe', 'df', pd.DataFrame}
+CSV_FORMATS = {'csv'}
+TSV_FORMATS = {'tsv'}
+JSON_FORMATS = {'json'}
+JSONL_FORMATS = {'jsonl'}
+EXCEL_FORMATS = {'excel'}
+PARQUET_FORMATS = {'parquet'}
+PICKLE_FORMATS = {'pickle'}
+FEATHER_FORMATS = {'feather'}
+FWF_FORMATS = {'fwf'}
+HTML_FORMATS = {'html'}
+ORC_FORMATS = {'orc'}
+SAS_FORMATS = {'sas'}
+SPSS_FORMATS = {'spss'}
+STATA_FORMATS = {'stata'}
+HDF5_FORMATS = {'hdf5', 'h5'}
+CACHEABLE_FORMATS = set.union(*(CSV_FORMATS, TSV_FORMATS,
+                                JSON_FORMATS, JSONL_FORMATS,
+                                EXCEL_FORMATS, PARQUET_FORMATS, PICKLE_FORMATS,
+                                FEATHER_FORMATS, FWF_FORMATS, HTML_FORMATS,
+                                ORC_FORMATS, SAS_FORMATS, SPSS_FORMATS,
+                                STATA_FORMATS))
+
+PANDAS_DF = pd
+
+
+def get_split_path(dataset_fp):
+    return os.path.splitext(dataset_fp)[0] + '.split.csv'
+
 
 def get_abs_path(data_csv_path, file_path):
     if data_csv_path is not None:
@@ -48,39 +86,100 @@ def load_csv(data_fp):
     return data
 
 
-def read_csv(data_fp, header=0, nrows=None, skiprows=None):
+def read_xsv(data_fp, df_lib=PANDAS_DF, separator=',', header=0, nrows=None, skiprows=None):
     """
     Helper method to read a csv file. Wraps around pd.read_csv to handle some
     exceptions. Can extend to cover cases as necessary
-    :param data_fp: path to the csv file
+    :param data_fp: path to the xsv file
+    :param df_lib: DataFrame library used to read in the CSV
+    :param separator: defaults separator to use for splitting
     :param header: header argument for pandas to read the csv
     :param nrows: number of rows to read from the csv, None means all
     :param skiprows: number of rows to skip from the csv, None means no skips
     :return: Pandas dataframe with the data
     """
-
-    separator = ','
     with open(data_fp, 'r', encoding="utf8") as csvfile:
         try:
             dialect = csv.Sniffer().sniff(csvfile.read(1024 * 100),
                                           delimiters=[',', '\t', '|'])
             separator = dialect.delimiter
         except csv.Error:
-            # Could not conclude the delimiter, defaulting to comma
+            # Could not conclude the delimiter, defaulting to user provided
             pass
 
     try:
-        df = pd.read_csv(data_fp, sep=separator, header=header,
-                         nrows=nrows, skiprows=skiprows)
+        df = df_lib.read_csv(data_fp, sep=separator, header=header,
+                             nrows=nrows, skiprows=skiprows)
     except ParserError:
         logger.warning('Failed to parse the CSV with pandas default way,'
                        ' trying \\ as escape character.')
-        df = pd.read_csv(data_fp, sep=separator, header=header, escapechar='\\',
-                         nrows=nrows, skiprows=skiprows)
+        df = df_lib.read_csv(data_fp, sep=separator, header=header,
+                             escapechar='\\',
+                             nrows=nrows, skiprows=skiprows)
 
     return df
 
 
+read_csv = functools.partial(read_xsv, separator=',')
+read_tsv = functools.partial(read_xsv, separator='\t')
+
+
+def read_json(data_fp, df_lib, normalize=False):
+    if normalize:
+        return df_lib.json_normalize(load_json(data_fp))
+    else:
+        return df_lib.read_json(data_fp)
+
+
+def read_jsonl(data_fp, df_lib):
+    return df_lib.read_json(data_fp, lines=True)
+
+
+def read_excel(data_fp, df_lib):
+    fp_split = os.path.splitext(data_fp)
+    if fp_split[1] == '.xls':
+        excel_engine = 'xlrd'
+    else:
+        excel_engine = 'openpyxl'
+    return df_lib.read_excel(data_fp, engine=excel_engine)
+
+
+def read_parquet(data_fp, df_lib):
+    return df_lib.read_parquet(data_fp)
+
+
+def read_pickle(data_fp, df_lib):
+    return df_lib.read_pickle(data_fp)
+
+
+def read_fwf(data_fp, df_lib):
+    return df_lib.read_fwf(data_fp)
+
+
+def read_feather(data_fp, df_lib):
+    return df_lib.read_feather(data_fp)
+
+
+def read_html(data_fp, df_lib):
+    return df_lib.read_html(data_fp)[0]
+
+
+def read_orc(data_fp, df_lib):
+    return df_lib.read_orc(data_fp)
+
+
+def read_sas(data_fp, df_lib):
+    return df_lib.read_sas(data_fp)
+
+
+def read_spss(data_fp, df_lib):
+    return df_lib.read_spss(data_fp)
+
+
+def read_stata(data_fp, df_lib):
+    return df_lib.read_stata(data_fp)
+
+
 def save_csv(data_fp, data):
     with open(data_fp, 'w', encoding='utf-8') as csv_file:
         writer = csv.writer(csv_file)
@@ -96,7 +195,6 @@ def csv_contains_column(data_fp, column_name):
 
 
 def load_json(data_fp):
-    data = []
     with open(data_fp, 'r') as input_file:
         data = json.load(input_file)
     return data
@@ -108,36 +206,49 @@ def save_json(data_fp, data, sort_keys=True, indent=4):
                   indent=indent)
 
 
-# to be tested
-# also, when loading an hdf5 file
-# most of the times you don't want
-# to put everything in memory
-# like this function does
-# it's jsut for convenience for relatively small datasets
-def load_hdf5(data_fp):
-    data = {}
-    with h5py.File(data_fp, 'r') as h5_file:
-        for key in h5_file.keys():
-            data[key] = h5_file[key][()]
-    return data
+def to_numpy_dataset(df):
+    dataset = {}
+    for col in df.columns:
+        dataset[col] = np.stack(df[col].to_numpy())
+    return dataset
 
 
-# def save_hdf5(data_fp: str, data: Dict[str, object]):
-def save_hdf5(data_fp, data, metadata=None):
-    if metadata is None:
-        metadata = {}
+def from_numpy_dataset(dataset):
+    col_mapping = {}
+    for k, v in dataset.items():
+        if len(v.shape) > 1:
+            # unstacking, needed for ndarrays of dimension 2 and more
+            *vals, = v
+        else:
+            # not unstacking. Needed because otherwise pandas casts types
+            # the way it wants, like converting a list of float32 scalats
+            # to a column of float64
+            vals = v
+        col_mapping[k] = vals
+    return pd.DataFrame.from_dict(col_mapping)
+
+
+def save_hdf5(data_fp, data):
     mode = 'w'
     if os.path.isfile(data_fp):
         mode = 'r+'
+
+    numpy_dataset = to_numpy_dataset(data)
     with h5py.File(data_fp, mode) as h5_file:
-        for key, value in data.items():
-            dataset = h5_file.create_dataset(key, data=value)
-            if key in metadata:
-                if 'in_memory' in metadata[key]['preprocessing']:
-                    if metadata[key]['preprocessing']['in_memory']:
-                        dataset.attrs['in_memory'] = True
-                    else:
-                        dataset.attrs['in_memory'] = False
+        h5_file.create_dataset(HDF5_COLUMNS_KEY, data=np.array(data.columns.values, dtype='S'))
+        for column in data.columns:
+            h5_file.create_dataset(column, data=numpy_dataset[column])
+
+
+def load_hdf5(data_fp):
+    hdf5_data = h5py.File(data_fp, 'r')
+    columns = [s.decode('utf-8') for s in hdf5_data[HDF5_COLUMNS_KEY][()].tolist()]
+
+    numpy_dataset = {}
+    for column in columns:
+        numpy_dataset[column] = hdf5_data[column][()]
+
+    return from_numpy_dataset(numpy_dataset)
 
 
 def load_object(object_fp):
@@ -195,7 +306,8 @@ def load_pretrained_embeddings(embeddings_path, vocab):
             embeddings_vectors.append(embeddings[word])
         else:
             embeddings_vectors.append(
-                avg_embedding + np.random.uniform(-0.01, 0.01, embeddings_size))
+                avg_embedding + np.random.uniform(-0.01, 0.01, embeddings_size)
+            )
     embeddings_matrix = np.stack(embeddings_vectors)
 
     # let's help the garbage collector free some memory
@@ -204,6 +316,7 @@ def load_pretrained_embeddings(embeddings_path, vocab):
     return embeddings_matrix
 
 
+@functools.lru_cache(1)
 def load_glove(file_path):
     logger.info('  Loading Glove format file {}'.format(file_path))
     embeddings = {}
@@ -276,37 +389,18 @@ def shuffle_dict_unison_inplace(np_dict, random_state=None):
     return recon
 
 
-def shuffle_inplace(np_dict):
-    if len(np_dict) == 0:
-        return
-    size = np_dict[next(iter(np_dict))].shape[0]
-    for k in np_dict:
-        if np_dict[k].shape[0] != size:
-            raise ValueError(
-                'Invalid: dictionary contains variable length arrays')
-
-    p = np.random.permutation(size)
-
-    for k in np_dict:
-        np_dict[k] = np_dict[k][p]
-
-
-def split_dataset_tvt(dataset, split):
-    if SPLIT in dataset:
-        del dataset[SPLIT]
-    training_set = split_dataset(dataset, split, value_to_split=0)
-    validation_set = split_dataset(dataset, split, value_to_split=1)
-    test_set = split_dataset(dataset, split, value_to_split=2)
+def split_dataset_ttv(dataset, split):
+    training_set = split_dataset(dataset, split, 0)
+    validation_set = split_dataset(dataset, split, 1)
+    test_set = split_dataset(dataset, split, 2)
     return training_set, test_set, validation_set
 
 
 def split_dataset(dataset, split, value_to_split=0):
-    splitted_dataset = {}
-    for key in dataset:
-        splitted_dataset[key] = dataset[key][split == value_to_split]
-        if len(splitted_dataset[key]) == 0:
-            return None
-    return splitted_dataset
+    split_df = dataset[dataset[split] == value_to_split]
+    if len(split_df) == 0:
+        return None
+    return split_df.reset_index()
 
 
 def collapse_rare_labels(labels, labels_limit):
@@ -320,7 +414,7 @@ def class_counts(dataset, labels_field):
 
 
 def text_feature_data_field(text_feature):
-    return text_feature['name'] + '_' + text_feature['level']
+    return text_feature[PROC_COLUMN] + '_' + text_feature['level']
 
 
 def load_from_file(file_name, field=None, dtype=int, ground_truth_split=2):
@@ -337,11 +431,9 @@ def load_from_file(file_name, field=None, dtype=int, ground_truth_split=2):
     :return: Experiment data as array
     """
     if file_name.endswith('.hdf5') and field is not None:
-        hdf5_data = h5py.File(file_name, 'r')
-        split = hdf5_data[SPLIT][()]
-        column = hdf5_data[field][()]
-        hdf5_data.close()
-        array = column[split == ground_truth_split]  # ground truth
+        dataset = pd.read_hdf(file_name, key=HDF5_COLUMNS_KEY)
+        column = dataset[field]
+        array = column[dataset[SPLIT] == ground_truth_split].values  # ground truth
     elif file_name.endswith('.npy'):
         array = np.load(file_name)
     elif file_name.endswith('.csv'):
@@ -362,9 +454,10 @@ def replace_file_extension(file_path, extension):
     """
     if file_path is None:
         return None
-    if '.' in extension:
+    extension = extension.strip()
+    if extension.startswith('.'):
         # Handle the case if the user calls with '.hdf5' instead of 'hdf5'
-        extension = extension.replace('.', '').strip()
+        extension = extension[1:]
 
     return os.path.splitext(file_path)[0] + '.' + extension
 
@@ -412,18 +505,29 @@ def add_sequence_feature_column(df, col_name, seq_length):
 def override_in_memory_flag(input_features, override_value):
     num_overrides = 0
     for feature in input_features:
-        if 'preprocessing' in feature:
-            if 'in_memory' in feature['preprocessing']:
-                feature['preprocessing']['in_memory'] = override_value
+        if PREPROCESSING in feature:
+            if 'in_memory' in feature[PREPROCESSING]:
+                feature[PREPROCESSING]['in_memory'] = override_value
                 num_overrides += 1
     return num_overrides
 
 
+def normalize_numpy(obj):
+    if isinstance(obj, np.integer):
+        return int(obj)
+    elif isinstance(obj, np.floating):
+        return float(obj)
+    elif isinstance(obj, np.ndarray):
+        return normalize_numpy(obj.tolist())
+    elif isinstance(obj, list):
+        return [normalize_numpy(v) for v in obj]
+    else:
+        return obj
+
+
 class NumpyEncoder(json.JSONEncoder):
     def default(self, obj):
-        if isinstance(obj, set):
-            return list(obj)
-        elif isinstance(obj, tuple):
+        if isinstance(obj, (set, tuple)):
             return list(obj)
         elif isinstance(obj, np.integer):
             return int(obj)
@@ -465,3 +569,124 @@ def get_path_size(
                     total_size += os.path.getsize(filepath)
 
     return total_size
+
+
+def clear_data_cache():
+    """Clears any cached data objects (e.g., embeddings)"""
+    load_glove.cache_clear()
+
+
+def figure_data_format_dataset(dataset):
+    if isinstance(dataset, pd.DataFrame):
+        return pd.DataFrame
+    elif isinstance(dataset, dict):
+        return dict
+    elif isinstance(dataset, str):
+        dataset = dataset.lower()
+        if dataset.endswith('.csv'):
+            return 'csv'
+        elif dataset.endswith('.tsv'):
+            return 'tsv'
+        elif dataset.endswith('.json'):
+            return 'json'
+        elif dataset.endswith('.jsonl'):
+            return 'jsonl'
+        elif (dataset.endswith('.xls') or dataset.endswith('.xlsx') or
+              dataset.endswith('.xlsm') or dataset.endswith('.xlsb') or
+              dataset.endswith('.odf') or dataset.endswith('.ods') or
+              dataset.endswith('.odt')):
+            return 'excel'
+        elif dataset.endswith('.parquet'):
+            return 'parquet'
+        elif dataset.endswith('.pickle') or dataset.endswith('.p'):
+            return 'pickle'
+        elif dataset.endswith('.feather'):
+            return 'feather'
+        elif dataset.endswith('.fwf'):
+            return 'fwf'
+        elif dataset.endswith('.html'):
+            return 'html'
+        elif dataset.endswith('.orc'):
+            return 'orc'
+        elif dataset.endswith('.sas'):
+            return 'sas'
+        elif dataset.endswith('.spss'):
+            return 'spss'
+        elif dataset.endswith('.dta') or dataset.endswith('.stata'):
+            return 'stata'
+        elif dataset.endswith('.h5') or dataset.endswith('.hdf5'):
+            return 'hdf5'
+        else:
+            raise ValueError(
+                "Dataset path string {} "
+                "does not contain a valid extension".format(dataset)
+            )
+    else:
+        raise ValueError(
+            "Cannot figure out the format of dataset {}".format(dataset)
+        )
+
+
+def figure_data_format(
+        dataset=None, training_set=None, validation_set=None, test_set=None
+):
+    if dataset is not None:
+        data_format = figure_data_format_dataset(dataset)
+    elif training_set is not None:
+        data_formats = [figure_data_format_dataset(training_set)]
+        if validation_set is not None:
+            data_formats.append(figure_data_format_dataset(validation_set))
+        if test_set is not None:
+            data_formats.append(figure_data_format_dataset(test_set))
+        data_formats_set = set(data_formats)
+        if len(data_formats_set) > 1:
+            error_message = "Datasets have different formats. Training: "
+            error_message += str(data_formats[0])
+            if validation_set:
+                error_message = ", Validation: "
+                error_message += str(data_formats[1])
+            if test_set:
+                error_message = ", Test: "
+                error_message += str(data_formats[-1])
+            raise ValueError(error_message)
+        else:
+            data_format = next(iter(data_formats_set))
+    else:
+        raise ValueError(
+            "At least one between dataset and training_set must be not None"
+        )
+    return data_format
+
+
+def is_model_dir(path: str) -> bool:
+    hyperparameters_fn = os.path.join(path, MODEL_HYPERPARAMETERS_FILE_NAME)
+    ts_metadata_fn = os.path.join(path, TRAIN_SET_METADATA_FILE_NAME)
+    is_model_dir = False
+    if (os.path.isdir(path)
+            and os.path.isfile(hyperparameters_fn)
+            and os.path.isfile(ts_metadata_fn)):
+        weights_files_count = 0
+        for file_name in os.listdir(path):
+            if file_name.startswith(MODEL_WEIGHTS_FILE_NAME):
+                weights_files_count += 1
+        if weights_files_count >= 2:
+            is_model_dir = True
+    return is_model_dir
+
+
+external_data_reader_registry = {
+    **{fmt: read_csv for fmt in CSV_FORMATS},
+    **{fmt: read_tsv for fmt in TSV_FORMATS},
+    **{fmt: read_json for fmt in JSON_FORMATS},
+    **{fmt: read_jsonl for fmt in JSONL_FORMATS},
+    **{fmt: read_excel for fmt in EXCEL_FORMATS},
+    **{fmt: read_parquet for fmt in PARQUET_FORMATS},
+    **{fmt: read_pickle for fmt in PICKLE_FORMATS},
+    **{fmt: read_fwf for fmt in FWF_FORMATS},
+    **{fmt: read_feather for fmt in FEATHER_FORMATS},
+    **{fmt: read_html for fmt in HTML_FORMATS},
+    **{fmt: read_orc for fmt in ORC_FORMATS},
+    **{fmt: read_sas for fmt in SAS_FORMATS},
+    **{fmt: read_spss for fmt in SPSS_FORMATS},
+    **{fmt: read_stata for fmt in STATA_FORMATS}
+}
