@@ -22,18 +22,22 @@ from collections import Counter
 
 import numpy as np
 
+from ludwig.data.dataframe.pandas import PANDAS
 from ludwig.utils.math_utils import int_type
-from ludwig.utils.misc import get_from_registry
+from ludwig.utils.misc_utils import get_from_registry
 from ludwig.utils.nlp_utils import load_nlp_pipeline, process_text
 
 UNKNOWN_SYMBOL = '<UNK>'
 PADDING_SYMBOL = '<PAD>'
+PADDING_IDX = 0
 
 SPLIT_REGEX = re.compile(r'\s+')
 SPACE_PUNCTUATION_REGEX = re.compile(r'\w+|[^\w\s]')
 COMMA_REGEX = re.compile(r'\s*,\s*')
 UNDERSCORE_REGEX = re.compile(r'\s*_\s*')
 
+BOOL_TRUE_STRS = {'yes', 'y', 'true', 't', '1'}
+
 
 def make_safe_filename(s):
     def safe_char(c):
@@ -51,7 +55,7 @@ def strip_accents(s):
 
 
 def str2bool(v):
-    return v.lower() in ('yes', 'true', 't', '1')
+    return str(v).lower() in BOOL_TRUE_STRS
 
 
 def match_replace(string_to_match, list_regex):
@@ -82,7 +86,7 @@ def load_vocabulary(vocab_file):
                 line = line.split(' ')[0]
             vocabulary.append(line)
         return vocabulary
-        #return [line.strip() for line in f]
+        # return [line.strip() for line in f]
 
 
 def create_vocabulary(
@@ -94,38 +98,63 @@ def create_vocabulary(
         num_most_frequent=None,
         vocab_file=None,
         unknown_symbol=UNKNOWN_SYMBOL,
-        padding_symbol=PADDING_SYMBOL
+        padding_symbol=PADDING_SYMBOL,
+        pretrained_model_name_or_path=None,
+        processor=PANDAS,
 ):
     vocab = None
-    max_line_length = 0
-    unit_counts = Counter()
-
-    if tokenizer_type == 'bert':
-        vocab = load_vocabulary(vocab_file)
-        add_unknown = False
-        add_padding = False
-    elif vocab_file is not None:
-        vocab = load_vocabulary(vocab_file)
 
     tokenizer = get_from_registry(
         tokenizer_type,
         tokenizer_registry
-    )(vocab_file=vocab_file)
-    for line in data:
-        processed_line = tokenizer(line.lower() if lowercase else line)
-        unit_counts.update(processed_line)
-        max_line_length = max(max_line_length, len(processed_line))
+    )(
+        vocab_file=vocab_file,
+        pretrained_model_name_or_path=pretrained_model_name_or_path,
+    )
+
+    if tokenizer_type == 'hf_tokenizer':
+        try:
+            vocab = tokenizer.tokenizer.get_vocab()
+            vocab = list(vocab.keys())
+        except NotImplementedError:
+            vocab = []
+            for idx in range(tokenizer.tokenizer.vocab_size):
+                vocab.append(tokenizer.tokenizer._convert_id_to_token(idx))
+            vocab += tokenizer.tokenizer.added_tokens_encoder.keys()
+
+        pad_token = tokenizer.tokenizer.pad_token
+        unk_token = tokenizer.tokenizer.unk_token
+
+        if pad_token is None:
+            vocab = vocab + [padding_symbol]
+        else:
+            padding_symbol = pad_token
+
+        if unk_token is None:
+            vocab = vocab + [unknown_symbol]
+        else:
+            unknown_symbol = unk_token
+
+    elif vocab_file is not None:
+        vocab = load_vocabulary(vocab_file)
+
+    processed_lines = data.map(lambda line: tokenizer(line.lower() if lowercase else line))
+    processed_counts = processed_lines.explode().value_counts(sort=False)
+    processed_counts = processor.compute(processed_counts)
+    unit_counts = Counter(dict(processed_counts))
+    max_line_length = processor.compute(processed_lines.map(len).max())
 
     if vocab is None:
         vocab = [unit for unit, count in
                  unit_counts.most_common(num_most_frequent)]
 
     vocab_set = set(vocab)
-    if add_unknown:
+
+    if add_unknown and tokenizer_type != 'hf_tokenizer':
         if unknown_symbol in vocab_set:
             vocab.remove(unknown_symbol)
         vocab = [unknown_symbol] + vocab
-    if add_padding:
+    if add_padding and tokenizer_type != 'hf_tokenizer':
         if padding_symbol in vocab_set:
             vocab.remove(padding_symbol)
         vocab = [padding_symbol] + vocab
@@ -134,15 +163,21 @@ def create_vocabulary(
     str2freq = {unit: unit_counts.get(unit) if unit in unit_counts else 0 for
                 unit in vocab}
 
-    return vocab, str2idx, str2freq, max_line_length
+    pad_idx = None
+    if padding_symbol in str2idx.keys():
+        pad_idx = str2idx[padding_symbol]
+
+    return vocab, str2idx, str2freq, max_line_length, pad_idx, padding_symbol, unknown_symbol
 
 
 def get_sequence_vector(sequence, tokenizer_type, unit_to_id, lowercase=True):
     tokenizer = get_from_registry(tokenizer_type, tokenizer_registry)()
+
     format_dtype = int_type(len(unit_to_id) - 1)
     return _get_sequence_vector(
         sequence,
         tokenizer,
+        tokenizer_type,
         format_dtype,
         unit_to_id,
         lowercase=lowercase
@@ -152,6 +187,7 @@ def get_sequence_vector(sequence, tokenizer_type, unit_to_id, lowercase=True):
 def _get_sequence_vector(
         sequence,
         tokenizer,
+        tokenizer_type,
         format_dtype,
         unit_to_id,
         lowercase=True,
@@ -160,13 +196,17 @@ def _get_sequence_vector(
     unit_sequence = tokenizer(
         sequence.lower() if lowercase else sequence
     )
+
     unit_indices_vector = np.empty(len(unit_sequence), dtype=format_dtype)
     for i in range(len(unit_sequence)):
         curr_unit = unit_sequence[i]
-        if curr_unit in unit_to_id:
-            unit_indices_vector[i] = unit_to_id[curr_unit]
+        if tokenizer_type == 'hf_tokenizer':
+            unit_indices_vector[i] = curr_unit
         else:
-            unit_indices_vector[i] = unit_to_id[unknown_symbol]
+            if curr_unit in unit_to_id:
+                unit_indices_vector[i] = unit_to_id[curr_unit]
+            else:
+                unit_indices_vector[i] = unit_to_id[unknown_symbol]
     return unit_indices_vector
 
 
@@ -180,42 +220,46 @@ def build_sequence_matrix(
         unknown_symbol=UNKNOWN_SYMBOL,
         lowercase=True,
         tokenizer_vocab_file=None,
+        pretrained_model_name_or_path=None,
+        processor=PANDAS,
 ):
     tokenizer = get_from_registry(tokenizer_type, tokenizer_registry)(
-        vocab_file=tokenizer_vocab_file
+        vocab_file=tokenizer_vocab_file,
+        pretrained_model_name_or_path=pretrained_model_name_or_path,
     )
+
     format_dtype = int_type(len(inverse_vocabulary) - 1)
 
-    max_length = 0
-    unit_vectors = []
-    for sequence in sequences:
-        unit_indices_vector = _get_sequence_vector(
-            sequence,
-            tokenizer,
-            format_dtype,
-            inverse_vocabulary,
-            lowercase=lowercase,
-            unknown_symbol=unknown_symbol
-        )
-        unit_vectors.append(unit_indices_vector)
-        if len(unit_indices_vector) > max_length:
-            max_length = len(unit_indices_vector)
+    unit_vectors = sequences.map(lambda sequence: _get_sequence_vector(
+        sequence,
+        tokenizer,
+        tokenizer_type,
+        format_dtype,
+        inverse_vocabulary,
+        lowercase=lowercase,
+        unknown_symbol=unknown_symbol
+    ))
 
+    max_length = processor.compute(unit_vectors.map(len).max())
     if max_length < length_limit:
         logging.debug('max length of {0}: {1} < limit: {2}'.format(
             format, max_length, length_limit
         ))
     max_length = length_limit
-    sequence_matrix = np.full((len(sequences), max_length),
-                              inverse_vocabulary[padding_symbol],
-                              dtype=format_dtype)
-    for i, vector in enumerate(unit_vectors):
+
+    def pad(vector):
+        sequence = np.full((max_length,),
+                           inverse_vocabulary[padding_symbol],
+                           dtype=format_dtype)
         limit = min(vector.shape[0], max_length)
         if padding == 'right':
-            sequence_matrix[i, :limit] = vector[:limit]
+            sequence[:limit] = vector[:limit]
         else:  # if padding == 'left
-            sequence_matrix[i, max_length - limit:] = vector[:limit]
-    return sequence_matrix
+            sequence[max_length - limit:] = vector[:limit]
+        return sequence
+
+    padded = processor.map_objects(unit_vectors, pad)
+    return padded
 
 
 class BaseTokenizer:
@@ -746,6 +790,7 @@ class NorwegianLemmatizeRemoveStopwordsFilterTokenizer(BaseTokenizer):
             filter_stopwords=True
         )
 
+
 class LithuanianTokenizer(BaseTokenizer):
     def __call__(self, text):
         return process_text(text, load_nlp_pipeline('lt'))
@@ -797,13 +842,272 @@ class LithuanianLemmatizeRemoveStopwordsFilterTokenizer(BaseTokenizer):
             filter_stopwords=True
         )
 
+
+class DanishTokenizer(BaseTokenizer):
+    def __call__(self, text):
+        return process_text(text, load_nlp_pipeline('da'))
+
+
+class DanishFilterTokenizer(BaseTokenizer):
+    def __call__(self, text):
+        return process_text(
+            text,
+            load_nlp_pipeline('da'),
+            filter_numbers=True,
+            filter_punctuation=True,
+            filter_short_tokens=True
+        )
+
+
+class DanishRemoveStopwordsTokenizer(BaseTokenizer):
+    def __call__(self, text):
+        return process_text(
+            text,
+            load_nlp_pipeline('da'),
+            filter_stopwords=True
+        )
+
+
+class DanishLemmatizeTokenizer(BaseTokenizer):
+    def __call__(self, text):
+        return process_text(text, load_nlp_pipeline('da'), return_lemma=True)
+
+
+class DanishLemmatizeFilterTokenizer(BaseTokenizer):
+    def __call__(self, text):
+        return process_text(
+            text,
+            load_nlp_pipeline('da'),
+            return_lemma=True,
+            filter_numbers=True,
+            filter_punctuation=True,
+            filter_short_tokens=True
+        )
+
+
+class DanishLemmatizeRemoveStopwordsFilterTokenizer(BaseTokenizer):
+    def __call__(self, text):
+        return process_text(
+            text,
+            load_nlp_pipeline('da'),
+            return_lemma=True,
+            filter_stopwords=True
+        )
+
+
+class PolishTokenizer(BaseTokenizer):
+    def __call__(self, text):
+        return process_text(text, load_nlp_pipeline('pl'))
+
+
+class PolishFilterTokenizer(BaseTokenizer):
+    def __call__(self, text):
+        return process_text(
+            text,
+            load_nlp_pipeline('pl'),
+            filter_numbers=True,
+            filter_punctuation=True,
+            filter_short_tokens=True
+        )
+
+
+class PolishRemoveStopwordsTokenizer(BaseTokenizer):
+    def __call__(self, text):
+        return process_text(
+            text,
+            load_nlp_pipeline('pl'),
+            filter_stopwords=True
+        )
+
+
+class PolishLemmatizeTokenizer(BaseTokenizer):
+    def __call__(self, text):
+        return process_text(text, load_nlp_pipeline('pl'), return_lemma=True)
+
+
+class PolishLemmatizeFilterTokenizer(BaseTokenizer):
+    def __call__(self, text):
+        return process_text(
+            text,
+            load_nlp_pipeline('pl'),
+            return_lemma=True,
+            filter_numbers=True,
+            filter_punctuation=True,
+            filter_short_tokens=True
+        )
+
+
+class PolishLemmatizeRemoveStopwordsFilterTokenizer(BaseTokenizer):
+    def __call__(self, text):
+        return process_text(
+            text,
+            load_nlp_pipeline('pl'),
+            return_lemma=True,
+            filter_stopwords=True
+        )
+
+
+class RomanianTokenizer(BaseTokenizer):
+    def __call__(self, text):
+        return process_text(text, load_nlp_pipeline('ro'))
+
+
+class RomanianFilterTokenizer(BaseTokenizer):
+    def __call__(self, text):
+        return process_text(
+            text,
+            load_nlp_pipeline('ro'),
+            filter_numbers=True,
+            filter_punctuation=True,
+            filter_short_tokens=True
+        )
+
+
+class RomanianRemoveStopwordsTokenizer(BaseTokenizer):
+    def __call__(self, text):
+        return process_text(
+            text,
+            load_nlp_pipeline('ro'),
+            filter_stopwords=True
+        )
+
+
+class RomanianLemmatizeTokenizer(BaseTokenizer):
+    def __call__(self, text):
+        return process_text(text, load_nlp_pipeline('ro'), return_lemma=True)
+
+
+class RomanianLemmatizeFilterTokenizer(BaseTokenizer):
+    def __call__(self, text):
+        return process_text(
+            text,
+            load_nlp_pipeline('ro'),
+            return_lemma=True,
+            filter_numbers=True,
+            filter_punctuation=True,
+            filter_short_tokens=True
+        )
+
+
+class RomanianLemmatizeRemoveStopwordsFilterTokenizer(BaseTokenizer):
+    def __call__(self, text):
+        return process_text(
+            text,
+            load_nlp_pipeline('ro'),
+            return_lemma=True,
+            filter_stopwords=True
+        )
+
+
+class JapaneseTokenizer(BaseTokenizer):
+    def __call__(self, text):
+        return process_text(text, load_nlp_pipeline('jp'))
+
+
+class JapaneseFilterTokenizer(BaseTokenizer):
+    def __call__(self, text):
+        return process_text(
+            text,
+            load_nlp_pipeline('jp'),
+            filter_numbers=True,
+            filter_punctuation=True,
+            filter_short_tokens=True
+        )
+
+
+class JapaneseRemoveStopwordsTokenizer(BaseTokenizer):
+    def __call__(self, text):
+        return process_text(
+            text,
+            load_nlp_pipeline('jp'),
+            filter_stopwords=True
+        )
+
+
+class JapaneseLemmatizeTokenizer(BaseTokenizer):
+    def __call__(self, text):
+        return process_text(text, load_nlp_pipeline('jp'), return_lemma=True)
+
+
+class JapaneseLemmatizeFilterTokenizer(BaseTokenizer):
+    def __call__(self, text):
+        return process_text(
+            text,
+            load_nlp_pipeline('jp'),
+            return_lemma=True,
+            filter_numbers=True,
+            filter_punctuation=True,
+            filter_short_tokens=True
+        )
+
+
+class JapaneseLemmatizeRemoveStopwordsFilterTokenizer(BaseTokenizer):
+    def __call__(self, text):
+        return process_text(
+            text,
+            load_nlp_pipeline('jp'),
+            return_lemma=True,
+            filter_stopwords=True
+        )
+
+
+class ChineseTokenizer(BaseTokenizer):
+    def __call__(self, text):
+        return process_text(text, load_nlp_pipeline('zh'))
+
+
+class ChineseFilterTokenizer(BaseTokenizer):
+    def __call__(self, text):
+        return process_text(
+            text,
+            load_nlp_pipeline('zh'),
+            filter_numbers=True,
+            filter_punctuation=True,
+            filter_short_tokens=True
+        )
+
+
+class ChineseRemoveStopwordsTokenizer(BaseTokenizer):
+    def __call__(self, text):
+        return process_text(
+            text,
+            load_nlp_pipeline('zh'),
+            filter_stopwords=True
+        )
+
+
+class ChineseLemmatizeTokenizer(BaseTokenizer):
+    def __call__(self, text):
+        return process_text(text, load_nlp_pipeline('zh'), return_lemma=True)
+
+
+class ChineseLemmatizeFilterTokenizer(BaseTokenizer):
+    def __call__(self, text):
+        return process_text(
+            text,
+            load_nlp_pipeline('zh'),
+            return_lemma=True,
+            filter_numbers=True,
+            filter_punctuation=True,
+            filter_short_tokens=True
+        )
+
+
+class ChineseLemmatizeRemoveStopwordsFilterTokenizer(BaseTokenizer):
+    def __call__(self, text):
+        return process_text(
+            text,
+            load_nlp_pipeline('zh'),
+            return_lemma=True,
+            filter_stopwords=True
+        )
+
+
 class MultiTokenizer(BaseTokenizer):
     def __call__(self, text):
         return process_text(text, load_nlp_pipeline('xx'))
 
 
-
-
 class MultiFilterTokenizer(BaseTokenizer):
     def __call__(self, text):
         return process_text(
@@ -851,25 +1155,20 @@ class MultiLemmatizeRemoveStopwordsTokenizer(BaseTokenizer):
         )
 
 
-class BERTTokenizer(BaseTokenizer):
-    def __init__(self, vocab_file=None, **kwargs):
+class HFTokenizer(BaseTokenizer):
+    def __init__(self,
+                 pretrained_model_name_or_path,
+                 **kwargs
+                 ):
+        super().__init__()
+        from transformers import AutoTokenizer
 
-        if vocab_file is None:
-            raise ValueError(
-                'Vocabulary file is required to initialize BERT tokenizer'
-            )
-
-        try:
-            from bert.tokenization import FullTokenizer
-        except ImportError:
-            raise ValueError(
-                "Please install bert-tensorflow: pip install bert-tensorflow"
-            )
-
-        self.tokenizer = FullTokenizer(vocab_file)
+        self.tokenizer = AutoTokenizer.from_pretrained(
+            pretrained_model_name_or_path,
+        )
 
     def __call__(self, text):
-        return ['[CLS]'] + self.tokenizer.tokenize(text) + ['[SEP]']
+        return self.tokenizer.encode(text, truncation=True)
 
 
 tokenizer_registry = {
@@ -940,11 +1239,41 @@ tokenizer_registry = {
     'lithuanian_lemmatize': LithuanianLemmatizeTokenizer,
     'lithuanian_lemmatize_filter': LithuanianLemmatizeFilterTokenizer,
     'lithuanian_lemmatize_remove_stopwords': LithuanianLemmatizeRemoveStopwordsFilterTokenizer,
+    'danish_tokenize': DanishTokenizer,
+    'danish_tokenize_filter': DanishFilterTokenizer,
+    'danish_tokenize_remove_stopwords': DanishRemoveStopwordsTokenizer,
+    'danish_lemmatize': DanishLemmatizeTokenizer,
+    'danish_lemmatize_filter': DanishLemmatizeFilterTokenizer,
+    'danish_lemmatize_remove_stopwords': DanishLemmatizeRemoveStopwordsFilterTokenizer,
+    'polish_tokenize': PolishTokenizer,
+    'polish_tokenize_filter': PolishFilterTokenizer,
+    'polish_tokenize_remove_stopwords': PolishRemoveStopwordsTokenizer,
+    'polish_lemmatize': PolishLemmatizeTokenizer,
+    'polish_lemmatize_filter': PolishLemmatizeFilterTokenizer,
+    'polish_lemmatize_remove_stopwords': PolishLemmatizeRemoveStopwordsFilterTokenizer,
+    'romanian_tokenize': RomanianTokenizer,
+    'romanian_tokenize_filter': RomanianFilterTokenizer,
+    'romanian_tokenize_remove_stopwords': RomanianRemoveStopwordsTokenizer,
+    'romanian_lemmatize': RomanianLemmatizeTokenizer,
+    'romanian_lemmatize_filter': RomanianLemmatizeFilterTokenizer,
+    'romanian_lemmatize_remove_stopwords': RomanianLemmatizeRemoveStopwordsFilterTokenizer,
+    'japanese_tokenize': JapaneseTokenizer,
+    'japanese_tokenize_filter': JapaneseFilterTokenizer,
+    'japanese_tokenize_remove_stopwords': JapaneseRemoveStopwordsTokenizer,
+    'japanese_lemmatize': JapaneseLemmatizeTokenizer,
+    'japanese_lemmatize_filter': JapaneseLemmatizeFilterTokenizer,
+    'japanese_lemmatize_remove_stopwords': JapaneseLemmatizeRemoveStopwordsFilterTokenizer,
+    'chinese_tokenize': ChineseTokenizer,
+    'chinese_tokenize_filter': ChineseFilterTokenizer,
+    'chinese_tokenize_remove_stopwords': ChineseRemoveStopwordsTokenizer,
+    'chinese_lemmatize': ChineseLemmatizeTokenizer,
+    'chinese_lemmatize_filter': ChineseLemmatizeFilterTokenizer,
+    'chinese_lemmatize_remove_stopwords': ChineseLemmatizeRemoveStopwordsFilterTokenizer,
     'multi_tokenize': MultiTokenizer,
     'multi_tokenize_filter': MultiFilterTokenizer,
     'multi_tokenize_remove_stopwords': MultiRemoveStopwordsTokenizer,
     'multi_lemmatize': MultiLemmatizeTokenizer,
     'multi_lemmatize_filter': MultiLemmatizeFilterTokenizer,
     'multi_lemmatize_remove_stopwords': MultiLemmatizeRemoveStopwordsTokenizer,
-    'bert': BERTTokenizer
+    'hf_tokenizer': HFTokenizer
 }
